@@ -1,24 +1,18 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
-import operator
-import itertools
 import functools
 import datetime
 from collections import defaultdict
+from functools import partial
 
-from sql import Table, Null
-from sql.functions import Overlay, Position
-from sql.aggregate import Max
-from sql.operators import Concat
+from sql import Null
 
 from trytond.model import Workflow, ModelView, ModelSQL, fields
 from trytond.modules.company import CompanyReport
 from trytond.wizard import Wizard, StateTransition, StateView, Button
-from trytond import backend
 from trytond.pyson import Eval, If, Id, Bool
 from trytond.transaction import Transaction
 from trytond.pool import Pool, PoolMeta
-from trytond.tools import reduce_ids, grouped_slice
 
 __all__ = ['ShipmentIn', 'ShipmentInReturn',
     'ShipmentOut', 'ShipmentOutReturn',
@@ -122,11 +116,17 @@ class ShipmentIn(Workflow, ModelSQL, ModelView):
                 ('shipment', '=', None),
                 ('from_location', '=', Eval('supplier_location')),
                 ('state', '=', 'draft'),
-                ('to_location', '=', Eval('warehouse_input')),
+                If(Eval('warehouse_input') == Eval('warehouse_storage'),
+                    ('to_location', 'child_of',
+                        [Eval('warehouse_input', -1)], 'parent'),
+                    ('to_location', '=', Eval('warehouse_input'))),
                 ],
             domain=[
                 ('from_location', '=', Eval('supplier_location')),
-                ('to_location', '=', Eval('warehouse_input')),
+                If(Eval('warehouse_input') == Eval('warehouse_storage'),
+                    ('to_location', 'child_of',
+                        [Eval('warehouse_input', -1)], 'parent'),
+                    ('to_location', '=', Eval('warehouse_input'))),
                 ('company', '=', Eval('company')),
                 ],
             states={
@@ -134,7 +134,7 @@ class ShipmentIn(Workflow, ModelSQL, ModelView):
                     | ~Eval('warehouse') | ~Eval('supplier')),
                 },
             depends=['state', 'warehouse', 'supplier_location',
-                'warehouse_input', 'company']),
+                'warehouse_input', 'warehouse_storage', 'company']),
         'get_incoming_moves', setter='set_incoming_moves')
     inventory_moves = fields.Function(fields.One2Many('stock.move', 'shipment',
             'Inventory Moves',
@@ -148,6 +148,8 @@ class ShipmentIn(Workflow, ModelSQL, ModelView):
                 ],
             states={
                 'readonly': Eval('state').in_(['draft', 'done', 'cancel']),
+                'invisible': (
+                    Eval('warehouse_input') == Eval('warehouse_storage')),
                 },
             depends=['state', 'warehouse', 'warehouse_input',
                 'warehouse_storage', 'company']),
@@ -188,69 +190,25 @@ class ShipmentIn(Workflow, ModelSQL, ModelView):
         cls._buttons.update({
                 'cancel': {
                     'invisible': Eval('state').in_(['cancel', 'done']),
+                    'depends': ['state'],
                     },
                 'draft': {
                     'invisible': Eval('state') != 'cancel',
+                    'depends': ['state'],
                     },
                 'receive': {
                     'invisible': Eval('state') != 'draft',
+                    'depends': ['state'],
                     },
                 'done': {
                     'invisible': Eval('state') != 'received',
+                    'depends': ['state'],
                     },
                 })
 
     @classmethod
     def __register__(cls, module_name):
-        TableHandler = backend.get('TableHandler')
-        cursor = Transaction().connection.cursor()
-        model_data = Table('ir_model_data')
-        model = Table('ir_model')
-        model_field = Table('ir_model_field')
-        sql_table = cls.__table__()
-
-        # Migration from 1.2: packing renamed into shipment
-        cursor.execute(*model_data.update(
-                columns=[model_data.fs_id],
-                values=[Overlay(model_data.fs_id, 'shipment',
-                        Position('packing', model_data.fs_id),
-                        len('packing'))],
-                where=model_data.fs_id.like('%packing%')
-                & (model_data.module == module_name)))
-        cursor.execute(*model.update(
-                columns=[model.model],
-                values=[Overlay(model.model, 'shipment',
-                        Position('packing', model.model),
-                        len('packing'))],
-                where=model.model.like('%packing%')
-                & (model.module == module_name)))
-        cursor.execute(*model_field.update(
-                columns=[model_field.relation],
-                values=[Overlay(model_field.relation, 'shipment',
-                        Position('packing', model_field.relation),
-                        len('packing'))],
-                where=model_field.relation.like('%packing%')
-                & (model_field.module == module_name)))
-        cursor.execute(*model_field.update(
-                columns=[model_field.name],
-                values=[Overlay(model_field.name, 'shipment',
-                        Position('packing', model_field.name),
-                        len('packing'))],
-                where=model_field.name.like('%packing%')
-                & (model_field.module == module_name)))
-
-        old_table = 'stock_packing_in'
-        if TableHandler.table_exist(old_table):
-            TableHandler.table_rename(old_table, cls._table)
-        table = TableHandler(cls, module_name)
-        for field in ('create_uid', 'write_uid', 'contact_address',
-                'warehouse', 'supplier'):
-            table.drop_fk(field, table=old_table)
-        for field in ('code', 'reference'):
-            table.index_action(field, action='remove', table=old_table)
-
-        # Migration from 2.0:
-        created_company = table.column_exist('company')
+        table = cls.__table_handler__(module_name)
 
         # Migration from 3.8: rename code into number
         if table.column_exist('code'):
@@ -258,30 +216,8 @@ class ShipmentIn(Workflow, ModelSQL, ModelView):
 
         super(ShipmentIn, cls).__register__(module_name)
 
-        # Migration from 2.0:
-        Move = Pool().get('stock.move')
-        if (not created_company
-                and TableHandler.table_exist(Move._table)):
-            move = Move.__table__()
-            cursor.execute(*sql_table.join(move,
-                    condition=(Concat(cls.__name__ + ',', sql_table.id)
-                        == move.shipment)
-                    ).select(sql_table.id, Max(move.company),
-                    group_by=sql_table.id,
-                    order_by=Max(move.company)))
-            for company_id, values in itertools.groupby(cursor.fetchall(),
-                    operator.itemgetter(1)):
-                shipment_ids = [x[0] for x in values]
-                for sub_ids in grouped_slice(shipment_ids):
-                    red_sql = reduce_ids(sql_table.id, sub_ids)
-                    cursor.execute(*sql_table.update(
-                            columns=[sql_table.company],
-                            values=[company_id],
-                            where=red_sql))
-            table.not_null_action('company', action='add')
-
         # Add index on create_date
-        table = TableHandler(cls, module_name)
+        table = cls.__table_handler__(module_name)
         table.index_action('create_date', action='add')
 
     @staticmethod
@@ -338,6 +274,8 @@ class ShipmentIn(Workflow, ModelSQL, ModelView):
 
     def get_incoming_moves(self, name):
         moves = []
+        if self.warehouse_input == self.warehouse_storage:
+            return [m.id for m in self.moves]
         for move in self.moves:
             if move.to_location == self.warehouse_input:
                 moves.append(move.id)
@@ -398,7 +336,7 @@ class ShipmentIn(Workflow, ModelSQL, ModelView):
                     })
 
     def get_origins(self, name):
-        return ', '.join(set(itertools.ifilter(None,
+        return ', '.join(set(filter(None,
                     (m.origin_name for m in self.moves))))
 
     @classmethod
@@ -425,12 +363,13 @@ class ShipmentIn(Workflow, ModelSQL, ModelView):
     def copy(cls, shipments, default=None):
         if default is None:
             default = {}
-        default = default.copy()
-        default['inventory_moves'] = None
-        default['incoming_moves'] = None
-        default.setdefault('number')
-        default.setdefault('received_by')
-        default.setdefault('done_by')
+        else:
+            default = default.copy()
+        default.setdefault('inventory_moves', None)
+        default.setdefault('incoming_moves', None)
+        default.setdefault('number', None)
+        default.setdefault('received_by', None)
+        default.setdefault('done_by', None)
         return super(ShipmentIn, cls).copy(shipments, default=default)
 
     @classmethod
@@ -455,6 +394,9 @@ class ShipmentIn(Workflow, ModelSQL, ModelView):
     @classmethod
     def create_inventory_moves(cls, shipments):
         for shipment in shipments:
+            if shipment.warehouse_storage == shipment.warehouse_input:
+                # Do not create inventory moves
+                continue
             # Use moves instead of inventory_moves because save reset before
             # adding new records and as set_inventory_moves is just a proxy to
             # moves, it will reset also the incoming_moves
@@ -503,6 +445,12 @@ class ShipmentIn(Workflow, ModelSQL, ModelView):
         Move = Pool().get('stock.move')
         Move.do([m for s in shipments for m in s.incoming_moves])
         cls.create_inventory_moves(shipments)
+        # Set received state to allow done transition
+        cls.write(shipments, {'state': 'received'})
+        to_do = [s for s in shipments
+            if s.warehouse_storage == s.warehouse_input]
+        if to_do:
+            cls.done(to_do)
 
     @classmethod
     @ModelView.button
@@ -623,22 +571,28 @@ class ShipmentInReturn(Workflow, ModelSQL, ModelView):
         cls._buttons.update({
                 'cancel': {
                     'invisible': Eval('state').in_(['cancel', 'done']),
+                    'depends': ['state'],
                     },
                 'draft': {
                     'invisible': ~Eval('state').in_(['waiting', 'cancel']),
-                    'icon': If(Eval('state') == 'cancel', 'tryton-clear',
-                        'tryton-go-previous'),
+                    'icon': If(Eval('state') == 'cancel',
+                        'tryton-undo',
+                        'tryton-forward'),
+                    'depends': ['state'],
                     },
                 'wait': {
                     'invisible': ~Eval('state').in_(['assigned', 'draft']),
                     'icon': If(Eval('state') == 'assigned',
-                        'tryton-go-previous', 'tryton-go-next'),
+                        'tryton-back', 'tryton-forward'),
+                    'depends': ['state'],
                     },
                 'done': {
                     'invisible': Eval('state') != 'assigned',
+                    'depends': ['state'],
                     },
                 'assign_wizard': {
                     'invisible': Eval('state') != 'waiting',
+                    'depends': ['state'],
                     },
                 'assign_try': {},
                 'assign_force': {},
@@ -646,22 +600,7 @@ class ShipmentInReturn(Workflow, ModelSQL, ModelView):
 
     @classmethod
     def __register__(cls, module_name):
-        TableHandler = backend.get('TableHandler')
-        cursor = Transaction().connection.cursor()
-        sql_table = cls.__table__()
-        # Migration from 1.2: packing renamed into shipment
-        old_table = 'stock_packing_in_return'
-        if TableHandler.table_exist(old_table):
-            TableHandler.table_rename(old_table, cls._table)
-        table = TableHandler(cls, module_name)
-        for field in ('create_uid', 'write_uid', 'from_location',
-                'to_location'):
-            table.drop_fk(field, table=old_table)
-        for field in ('code', 'reference'):
-            table.index_action(field, action='remove', table=old_table)
-
-        # Migration from 2.0:
-        created_company = table.column_exist('company')
+        table = cls.__table_handler__(module_name)
 
         # Migration from 3.8: rename code into number
         if table.column_exist('code'):
@@ -669,30 +608,8 @@ class ShipmentInReturn(Workflow, ModelSQL, ModelView):
 
         super(ShipmentInReturn, cls).__register__(module_name)
 
-        # Migration from 2.0:
-        Move = Pool().get('stock.move')
-        if (not created_company
-                and TableHandler.table_exist(Move._table)):
-            move = Move.__table__()
-            cursor.execute(*sql_table.join(move,
-                    condition=(Concat(cls.__name__ + ',', sql_table.id)
-                        == move.shipment)
-                    ).select(sql_table.id, Max(move.company),
-                    group_by=sql_table.id,
-                    order_by=Max(move.company)))
-            for company_id, values in itertools.groupby(cursor.fetchall(),
-                    operator.itemgetter(1)):
-                shipment_ids = [x[0] for x in values]
-                for sub_ids in grouped_slice(shipment_ids):
-                    red_sql = reduce_ids(sql_table.id, sub_ids)
-                    cursor.execute(*sql_table.update(
-                            columns=[sql_table.company],
-                            values=[company_id],
-                            where=red_sql))
-            table.not_null_action('company', action='add')
-
         # Add index on create_date
-        table = TableHandler(cls, module_name)
+        table = cls.__table_handler__(module_name)
         table.index_action('create_date', action='add')
 
     @staticmethod
@@ -730,7 +647,7 @@ class ShipmentInReturn(Workflow, ModelSQL, ModelView):
                     })
 
     def get_origins(self, name):
-        return ', '.join(set(itertools.ifilter(None,
+        return ', '.join(set(filter(None,
                     (m.origin_name for m in self.moves))))
 
     @classmethod
@@ -769,10 +686,11 @@ class ShipmentInReturn(Workflow, ModelSQL, ModelView):
     def copy(cls, shipments, default=None):
         if default is None:
             default = {}
-        default = default.copy()
-        default.setdefault('number')
-        default.setdefault('assigned_by')
-        default.setdefault('done_by')
+        else:
+            default = default.copy()
+        default.setdefault('number', None)
+        default.setdefault('assigned_by', None)
+        default.setdefault('done_by', None)
         return super(ShipmentInReturn, cls).copy(shipments, default=default)
 
     @classmethod
@@ -834,16 +752,28 @@ class ShipmentInReturn(Workflow, ModelSQL, ModelView):
 
     @classmethod
     @ModelView.button
-    def assign_try(cls, shipments):
+    def assign_try(cls, shipments, with_childs=None):
         pool = Pool()
         Move = pool.get('stock.move')
-        to_assign = [m for s in shipments for m in s.moves
-            if m.assignation_required]
-        if Move.assign_try(to_assign, with_childs=False):
+        to_assign = defaultdict(list)
+        for shipment in shipments:
+            location_type = shipment.from_location.type
+            for move in shipment.moves:
+                if move.assignation_required:
+                    to_assign[location_type].append(move)
+        success = True
+        for location_type, moves in to_assign.items():
+            if with_childs is None:
+                _with_childs = location_type == 'view'
+            elif not with_childs and location_type == 'view':
+                _with_childs = True
+            else:
+                _with_childs = with_childs
+            if not Move.assign_try(moves, with_childs=_with_childs):
+                success = False
+        if success:
             cls.assign(shipments)
-            return True
-        else:
-            return False
+        return success
 
     @classmethod
     @ModelView.button
@@ -978,26 +908,32 @@ class ShipmentOut(Workflow, ModelSQL, ModelView):
         cls._buttons.update({
                 'cancel': {
                     'invisible': Eval('state').in_(['cancel', 'done']),
+                    'depends': ['state'],
                     },
                 'draft': {
                     'invisible': ~Eval('state').in_(['waiting', 'cancel']),
-                    'icon': If(Eval('state') == 'cancel', 'tryton-clear',
-                        'tryton-go-previous'),
+                    'icon': If(Eval('state') == 'cancel',
+                        'tryton-undo',
+                        'tryton-forward'),
+                    'depends': ['state'],
                     },
                 'wait': {
                     'invisible': ~Eval('state').in_(['assigned', 'waiting',
                             'draft']),
                     'icon': If(Eval('state') == 'assigned',
-                        'tryton-go-previous',
+                        'tryton-back',
                         If(Eval('state') == 'waiting',
                             'tryton-clear',
-                            'tryton-go-next')),
+                            'tryton-forward')),
+                    'depends': ['state'],
                     },
                 'pack': {
                     'invisible': If(Eval('warehouse_storage')
                         == Eval('warehouse_output'),
                         ~Eval('state').in_(['assigned', 'waiting']),
                         Eval('state') != 'assigned'),
+                    'depends': ['state', 'warehouse_storage',
+                        'warehouse_output'],
                     },
                 'done': {
                     'invisible': Eval('state') != 'packed',
@@ -1013,23 +949,7 @@ class ShipmentOut(Workflow, ModelSQL, ModelView):
 
     @classmethod
     def __register__(cls, module_name):
-        TableHandler = backend.get('TableHandler')
-        cursor = Transaction().connection.cursor()
-        sql_table = cls.__table__()
-        # Migration from 1.2: packing renamed into shipment
-        old_table = 'stock_packing_out'
-        if TableHandler.table_exist(old_table):
-            TableHandler.table_rename(old_table, cls._table)
-
-        table = TableHandler(cls, module_name)
-        for field in ('create_uid', 'write_uid', 'delivery_address',
-                'warehouse', 'customer'):
-            table.drop_fk(field, table=old_table)
-        for field in ('code', 'reference'):
-            table.index_action(field, action='remove', table=old_table)
-
-        # Migration from 2.0:
-        created_company = table.column_exist('company')
+        table = cls.__table_handler__(module_name)
 
         # Migration from 3.8: rename code into number
         if table.column_exist('code'):
@@ -1037,32 +957,6 @@ class ShipmentOut(Workflow, ModelSQL, ModelView):
 
         super(ShipmentOut, cls).__register__(module_name)
 
-        # Migration from 2.0:
-        Move = Pool().get('stock.move')
-        if (not created_company
-                and TableHandler.table_exist(Move._table)):
-            move = Move.__table__()
-            cursor.execute(*sql_table.join(move,
-                    condition=(Concat(cls.__name__ + ',', sql_table.id)
-                        == move.shipment)
-                    ).select(sql_table.id, Max(move.company),
-                    group_by=sql_table.id,
-                    order_by=Max(move.company)))
-            for company_id, values in itertools.groupby(cursor.fetchall(),
-                    operator.itemgetter(1)):
-                shipment_ids = [x[0] for x in values]
-                for sub_ids in grouped_slice(shipment_ids):
-                    red_sql = reduce_ids(sql_table.id, sub_ids)
-                    cursor.execute(*sql_table.update(
-                            columns=[sql_table.company],
-                            values=[company_id],
-                            where=red_sql))
-            table.not_null_action('company', action='add')
-
-        # Migration from 1.0 customer_location is no more used
-        table = TableHandler(cls, module_name)
-        table.drop_column('customer_location', exception=True)
-
         # Add index on create_date
         table.index_action('create_date', action='add')
 
@@ -1149,7 +1043,7 @@ class ShipmentOut(Workflow, ModelSQL, ModelView):
                 })
 
     def get_origins(self, name):
-        return ', '.join(set(itertools.ifilter(None,
+        return ', '.join(set(filter(None,
                     (m.origin_name for m in self.moves))))
 
     @classmethod
@@ -1339,7 +1233,7 @@ class ShipmentOut(Workflow, ModelSQL, ModelView):
         return self.planned_date, self.planned_date
 
     @classmethod
-    def _set_move_planned_date(self, shipments):
+    def _set_move_planned_date(cls, shipments):
         '''
         Set planned date of moves for the shipments
         '''
@@ -1390,13 +1284,14 @@ class ShipmentOut(Workflow, ModelSQL, ModelView):
     def copy(cls, shipments, default=None):
         if default is None:
             default = {}
-        default = default.copy()
-        default['inventory_moves'] = None
-        default['outgoing_moves'] = None
-        default.setdefault('number')
-        default.setdefault('assigned_by')
-        default.setdefault('packed_by')
-        default.setdefault('done_by')
+        else:
+            default = default.copy()
+        default.setdefault('inventory_moves', None)
+        default.setdefault('outgoing_moves', None)
+        default.setdefault('number', None)
+        default.setdefault('assigned_by', None)
+        default.setdefault('packed_by', None)
+        default.setdefault('done_by', None)
         return super(ShipmentOut, cls).copy(shipments, default=default)
 
     @classmethod
@@ -1502,7 +1397,10 @@ class ShipmentOutReturn(Workflow, ModelSQL, ModelView):
             'Incoming Moves',
             domain=[
                 ('from_location', '=', Eval('customer_location')),
-                ('to_location', '=', Eval('warehouse_input')),
+                If(Eval('warehouse_input') == Eval('warehouse_storage'),
+                    ('to_location', 'child_of',
+                        [Eval('warehouse_input', -1)], 'parent'),
+                    ('to_location', '=', Eval('warehouse_input'))),
                 ('company', '=', Eval('company')),
                 ],
             states={
@@ -1510,7 +1408,7 @@ class ShipmentOutReturn(Workflow, ModelSQL, ModelView):
                     | ~Eval('warehouse') | ~Eval('customer')),
                 },
             depends=['state', 'warehouse', 'customer', 'customer_location',
-                'warehouse_input', 'company']),
+                'warehouse_input', 'warehouse_storage', 'company']),
         'get_incoming_moves', setter='set_incoming_moves')
     inventory_moves = fields.Function(fields.One2Many('stock.move', 'shipment',
             'Inventory Moves',
@@ -1524,9 +1422,11 @@ class ShipmentOutReturn(Workflow, ModelSQL, ModelView):
                 ],
             states={
                 'readonly': Eval('state').in_(['draft', 'cancel', 'done']),
+                'invisible': (
+                    Eval('warehouse_input') == Eval('warehouse_storage')),
                 },
             depends=['state', 'warehouse', 'warehouse_input',
-                'warehouse_storage', 'company']),
+                'warehouse_storage', 'warehouse_input', 'company']),
         'get_inventory_moves', setter='set_inventory_moves')
     moves = fields.One2Many('stock.move', 'shipment', 'Moves',
         domain=[('company', '=', Eval('company'))], depends=['company'],
@@ -1561,37 +1461,25 @@ class ShipmentOutReturn(Workflow, ModelSQL, ModelView):
         cls._buttons.update({
                 'cancel': {
                     'invisible': Eval('state').in_(['cancel', 'done']),
+                    'depends': ['state'],
                     },
                 'draft': {
                     'invisible': Eval('state') != 'cancel',
+                    'depends': ['state'],
                     },
                 'receive': {
                     'invisible': Eval('state') != 'draft',
+                    'depends': ['state'],
                     },
                 'done': {
                     'invisible': Eval('state') != 'received',
+                    'depends': ['state'],
                     },
                 })
 
     @classmethod
     def __register__(cls, module_name):
-        TableHandler = backend.get('TableHandler')
-        cursor = Transaction().connection.cursor()
-        sql_table = cls.__table__()
-        # Migration from 1.2: packing renamed into shipment
-        old_table = 'stock_packing_out_return'
-        if TableHandler.table_exist(old_table):
-            TableHandler.table_rename(old_table, cls._table)
-
-        table = TableHandler(cls, module_name)
-        for field in ('create_uid', 'write_uid', 'delivery_address',
-                'warehouse', 'customer'):
-            table.drop_fk(field, table=old_table)
-        for field in ('code', 'reference'):
-            table.index_action(field, action='remove', table=old_table)
-
-        # Migration from 2.0:
-        created_company = table.column_exist('company')
+        table = cls.__table_handler__(module_name)
 
         # Migration from 3.8: rename code into number
         if table.column_exist('code'):
@@ -1599,30 +1487,8 @@ class ShipmentOutReturn(Workflow, ModelSQL, ModelView):
 
         super(ShipmentOutReturn, cls).__register__(module_name)
 
-        # Migration from 2.0:
-        Move = Pool().get('stock.move')
-        if (not created_company
-                and TableHandler.table_exist(Move._table)):
-            move = Move.__table__()
-            cursor.execute(*sql_table.join(move,
-                    condition=(Concat(cls.__name__ + ',', sql_table.id)
-                        == move.shipment)
-                    ).select(sql_table.id, Max(move.company),
-                    group_by=sql_table.id,
-                    order_by=Max(move.company)))
-            for company_id, values in itertools.groupby(cursor.fetchall(),
-                    operator.itemgetter(1)):
-                shipment_ids = [x[0] for x in values]
-                for sub_ids in grouped_slice(shipment_ids):
-                    red_sql = reduce_ids(sql_table.id, sub_ids)
-                    cursor.execute(*sql_table.update(
-                            columns=[sql_table.company],
-                            values=[company_id],
-                            where=red_sql))
-            table.not_null_action('company', action='add')
-
         # Add index on create_date
-        table = TableHandler(cls, module_name)
+        table = cls.__table_handler__(module_name)
         table.index_action('create_date', action='add')
 
     @staticmethod
@@ -1675,6 +1541,8 @@ class ShipmentOutReturn(Workflow, ModelSQL, ModelView):
 
     def get_incoming_moves(self, name):
         moves = []
+        if self.warehouse_input == self.warehouse_storage:
+            return [m.id for m in self.moves]
         for move in self.moves:
             if move.to_location == self.warehouse_input:
                 moves.append(move.id)
@@ -1730,7 +1598,7 @@ class ShipmentOutReturn(Workflow, ModelSQL, ModelView):
                     })
 
     def get_origins(self, name):
-        return ', '.join(set(itertools.ifilter(None,
+        return ', '.join(set(filter(None,
                     (m.origin_name for m in self.moves))))
 
     @classmethod
@@ -1759,11 +1627,11 @@ class ShipmentOutReturn(Workflow, ModelSQL, ModelView):
         if default is None:
             default = {}
         default = default.copy()
-        default['inventory_moves'] = None
-        default['incoming_moves'] = None
-        default.setdefault('number')
-        default.setdefault('received_by')
-        default.setdefault('done_by')
+        default.setdefault('inventory_moves', None)
+        default.setdefault('incoming_moves', None)
+        default.setdefault('number', None)
+        default.setdefault('received_by', None)
+        default.setdefault('done_by', None)
         return super(ShipmentOutReturn, cls).copy(shipments, default=default)
 
     @classmethod
@@ -1795,6 +1663,12 @@ class ShipmentOutReturn(Workflow, ModelSQL, ModelView):
         Move = Pool().get('stock.move')
         Move.do([m for s in shipments for m in s.incoming_moves])
         cls.create_inventory_moves(shipments)
+        # Set received state to allow done transition
+        cls.write(shipments, {'state': 'received'})
+        to_do = [s for s in shipments
+            if s.warehouse_storage == s.warehouse_input]
+        if to_do:
+            cls.done(to_do)
 
     @classmethod
     @ModelView.button
@@ -1839,6 +1713,9 @@ class ShipmentOutReturn(Workflow, ModelSQL, ModelView):
     @classmethod
     def create_inventory_moves(cls, shipments):
         for shipment in shipments:
+            if shipment.warehouse_storage == shipment.warehouse_input:
+                # Do not create inventory moves
+                continue
             # Use moves instead of inventory_moves because save reset before
             # adding new records and as set_inventory_moves is just a proxy to
             # moves, it will reset also the incoming_moves
@@ -1852,7 +1729,7 @@ class ShipmentOutReturn(Workflow, ModelSQL, ModelView):
 
 
 class AssignShipmentOutAssignFailed(ModelView):
-    'Assign Shipment Out'
+    'Assign Customer Shipment'
     __name__ = 'stock.shipment.out.assign.failed'
     inventory_moves = fields.Many2Many('stock.move', None, None,
         'Inventory Moves', readonly=True)
@@ -1868,12 +1745,12 @@ class AssignShipmentOutAssignFailed(ModelView):
 
 
 class AssignShipmentOut(Wizard):
-    'Assign Shipment Out'
+    'Assign Customer Shipment'
     __name__ = 'stock.shipment.out.assign'
     start = StateTransition()
     failed = StateView('stock.shipment.out.assign.failed',
         'stock.shipment_out_assign_failed_view_form', [
-            Button('Force Assign', 'force', 'tryton-go-next',
+            Button('Force Assign', 'force', 'tryton-forward',
                 states={
                     'invisible': ~Id('stock',
                         'group_stock_force_assignment').in_(
@@ -2076,37 +1953,43 @@ class ShipmentInternal(Workflow, ModelSQL, ModelView):
                 'cancel': {
                     'invisible': Eval('state').in_(
                         ['cancel', 'shipped', 'done']),
+                    'depends': ['state'],
                     },
                 'draft': {
                     'invisible': ~Eval('state').in_(
                         ['cancel', 'request', 'waiting']),
                     'icon': If(Eval('state') == 'cancel',
-                        'tryton-clear',
+                        'tryton-undo',
                         If(Eval('state') == 'request',
-                            'tryton-go-next',
-                            'tryton-go-previous')),
+                            'tryton-forward',
+                            'tryton-back')),
+                    'depends': ['state'],
                     },
                 'wait': {
                     'invisible': ~Eval('state').in_(['assigned', 'waiting',
                             'draft']),
                     'icon': If(Eval('state') == 'assigned',
-                        'tryton-go-previous',
+                        'tryton-back',
                         If(Eval('state') == 'waiting',
                             'tryton-clear',
-                            'tryton-go-next')),
+                            'tryton-forward')),
+                    'depends': ['state'],
                     },
                 'ship': {
                     'invisible': ((Eval('state') != 'assigned') |
                         ~Eval('transit_location')),
+                    'depends': ['state', 'transit_location'],
                     },
                 'done': {
                     'invisible': If(
                         ~Eval('transit_location'),
                         Eval('state') != 'assigned',
                         Eval('state') != 'shipped'),
+                    'depends': ['state', 'transit_location'],
                     },
                 'assign_wizard': {
                     'invisible': Eval('state') != 'waiting',
+                    'depends': ['state'],
                     },
                 'assign_try': {},
                 'assign_force': {},
@@ -2114,22 +1997,9 @@ class ShipmentInternal(Workflow, ModelSQL, ModelView):
 
     @classmethod
     def __register__(cls, module_name):
-        TableHandler = backend.get('TableHandler')
         cursor = Transaction().connection.cursor()
         sql_table = cls.__table__()
-        # Migration from 1.2: packing renamed into shipment
-        old_table = 'stock_packing_internal'
-        if TableHandler.table_exist(old_table):
-            TableHandler.table_rename(old_table, cls._table)
-        table = TableHandler(cls, module_name)
-        for field in ('create_uid', 'write_uid', 'from_location',
-                'to_location'):
-            table.drop_fk(field, table=old_table)
-        for field in ('code', 'reference'):
-            table.index_action(field, action='remove', table=old_table)
-
-        # Migration from 2.0:
-        created_company = table.column_exist('company')
+        table = cls.__table_handler__(module_name)
 
         # Migration from 3.8:
         if table.column_exist('code'):
@@ -2137,28 +2007,6 @@ class ShipmentInternal(Workflow, ModelSQL, ModelView):
 
         super(ShipmentInternal, cls).__register__(module_name)
 
-        # Migration from 2.0:
-        Move = Pool().get('stock.move')
-        if (not created_company
-                and TableHandler.table_exist(Move._table)):
-            move = Move.__table__()
-            cursor.execute(*sql_table.join(move,
-                    condition=(Concat(cls.__name__ + ',', sql_table.id)
-                        == move.shipment)
-                    ).select(sql_table.id, Max(move.company),
-                    group_by=sql_table.id,
-                    order_by=Max(move.company)))
-            for company_id, values in itertools.groupby(cursor.fetchall(),
-                    operator.itemgetter(1)):
-                shipment_ids = [x[0] for x in values]
-                for sub_ids in grouped_slice(shipment_ids):
-                    red_sql = reduce_ids(sql_table.id, sub_ids)
-                    cursor.execute(*sql_table.update(
-                            columns=[sql_table.company],
-                            values=[company_id],
-                            where=red_sql))
-            table.not_null_action('company', action='add')
-
         # Migration from 4.0: fill planned_start_date
         cursor = Transaction().connection.cursor()
         cursor.execute(*sql_table.update(
@@ -2168,7 +2016,7 @@ class ShipmentInternal(Workflow, ModelSQL, ModelView):
                 & (sql_table.planned_date != Null)))
 
         # Add index on create_date
-        table = TableHandler(cls, module_name)
+        table = cls.__table_handler__(module_name)
         table.index_action('create_date', action='add')
 
     @staticmethod
@@ -2259,29 +2107,34 @@ class ShipmentInternal(Workflow, ModelSQL, ModelView):
 
     @classmethod
     def copy(cls, shipments, default=None):
-        pool = Pool()
-        Move = pool.get('stock.move')
+        def shipment_field(data, name):
+            model, shipment_id = data['shipment'].split(',', 1)
+            assert model == cls.__name__
+            shipment_id = int(shipment_id)
+            shipment = id2shipments[shipment_id]
+            return getattr(shipment, name)
+
+        def outgoing_moves(data):
+            shipment = id2shipments[data['id']]
+            return shipment.outgoing_moves
+        id2shipments = {s.id: s for s in shipments}
 
         if default is None:
             default = {}
         else:
             default = default.copy()
-        default['outgoing_moves'] = None
-        default['incoming_moves'] = None
-        default['moves'] = None
-        default.setdefault('number')
-        default.setdefault('assigned_by')
-        default.setdefault('shipped_by')
-        default.setdefault('done_by')
-        copies = super(ShipmentInternal, cls).copy(shipments, default=default)
-        for shipment, copy in zip(shipments, copies):
-            Move.copy(shipment.outgoing_moves, default={
-                    'shipment': str(copy),
-                    'from_location': shipment.from_location.id,
-                    'to_location': shipment.to_location.id,
-                    'planned_date': shipment.planned_date,
-                    })
-        return copies
+        default.setdefault('moves', outgoing_moves)
+        default.setdefault('moves.from_location', partial(
+                shipment_field, name='from_location'))
+        default.setdefault('moves.to_location', partial(
+                shipment_field, name='to_location'))
+        default.setdefault('moves.planned_date', partial(
+                shipment_field, name='planned_date'))
+        default.setdefault('number', None)
+        default.setdefault('assigned_by', None)
+        default.setdefault('shipped_by', None)
+        default.setdefault('done_by', None)
+        return super().copy(shipments, default=default)
 
     @classmethod
     def _sync_moves(cls, shipments):
@@ -2452,8 +2305,7 @@ class ShipmentInternal(Workflow, ModelSQL, ModelView):
         cls.assign(shipments)
 
 
-class Address:
-    __metaclass__ = PoolMeta
+class Address(metaclass=PoolMeta):
     __name__ = 'party.address'
     delivery = fields.Boolean('Delivery')
 
@@ -2480,7 +2332,7 @@ class AssignShipmentInternal(Wizard):
     start = StateTransition()
     failed = StateView('stock.shipment.internal.assign.failed',
         'stock.shipment_internal_assign_failed_view_form', [
-            Button('Force Assign', 'force', 'tryton-go-next',
+            Button('Force Assign', 'force', 'tryton-forward',
                 states={
                     'invisible': ~Id('stock',
                         'group_stock_force_assignment').in_(
@@ -2528,7 +2380,7 @@ class AssignShipmentInReturn(Wizard):
     start = StateTransition()
     failed = StateView('stock.shipment.in.return.assign.failed',
         'stock.shipment_in_return_assign_failed_view_form', [
-            Button('Force Assign', 'force', 'tryton-go-next',
+            Button('Force Assign', 'force', 'tryton-forward',
                 states={
                     'invisible': ~Id('stock',
                         'group_stock_force_assignment').in_(
